/*
*  C Implementation: AVR ISP Programmer 
*
* Description:
*
*  A serial programmer  somewhat compatible with "butterfly" "avr109" and
*     "avr910" modes of avrdude
*
*
* Author: Koji Suzuki (C) 2007,2010
* ATTiny2313_bootloader Author:
*              Ernst Bachmann <e.bachmann@netandweb.de>, (C) 2006
*
* Copyright: See License.txt file that comes with this distribution
*
*/

#include <usb910conf.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>

#if defined(__CDC162__) 
#include <cdc162.h>
#else
#include <usbcdc.h>
#endif
#include <usb910.h>

#if !defined(NO_AVR910) && ( TX_SIZE < (32) )
#error "TX_SIZE is too small"
#endif

#ifdef JTAG_EXTENSION
#define JTAG_BUFSIZE	64
#endif

#define getc	usbcdc_getc
#define putc	usbcdc_putc
#define poll	usbcdc_poll

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#define swap(v) __asm__ volatile ("swap %0" : "=r" (v) : "0" (v) )
#define nop()   __asm__ volatile ("nop" )

#ifndef NO_AVR910
static void init_timeout() {
#if defined(__AVR_ATtiny861__) || defined(__AVR_ATtiny461__)
	TCCR0A = (1<<TCW0); // 16 bit mode
	TCCR0B = 5; // timer0 start @ clk/1024
#else
	TCCR1B = 5; // timer1 start @ clk/1024
#endif
}

static inline uint16_t get_time() {
#if defined(__AVR_ATtiny861__) || defined(__AVR_ATtiny461__)
	union {
		uint16_t word;
		struct {
			uint8_t lo;
			uint8_t hi;
		};
	} timer;
	timer.lo = TCNT0L;
	timer.hi = TCNT0H;
	return timer.word;
#else
	return TCNT1;
#endif
}

static int8_t to_flag;
static int16_t to_time;

static void set_timeout(uint16_t t) {
	to_time = t + get_time();
	to_flag = 1;
}

static int8_t check_timeout() {
	if (to_flag && (to_time - (int16_t)(get_time()) <= 0)) {
		to_flag = 0;
	}
	return !to_flag;
}

static void __delay(uint16_t t) {
	set_timeout(t);
	while (check_timeout()) {
		wdt_reset();
		poll();
	}
}

/* mdelay -- max 2.79 sec @12Mhz  */
static inline void mdelay(uint8_t ms) {
	__delay((uint16_t)ms * (F_CPU/1000) / 1024);
}

/* 
 * MAX SPI CLOCK = 12Mhz/3 /3 / 6 = 222 kHz @ F_CPU==12Mhz
 * MIN CPU CLOCK = 222 kHz * 4 = 888 kHz  < ( 1MHz = 8Mhz/8 )
 */
#define DELAY_BASE	(uint8_t)(F_CPU/4000000)
static void spi_clock(void) {
	_delay_loop_1(DELAY_BASE *1);
	bit_set(TGT_PORT, TGT_SCK);
	_delay_loop_1(DELAY_BASE *3);
	bit_clear(TGT_PORT, TGT_SCK);
	_delay_loop_1(DELAY_BASE *2 -2);
}

static uint8_t spi_write(uint8_t c) {
	uint8_t i,rx;
	i = 0x80;
	rx = 0;
	do {
		if (c & i) {
			bit_set(TGT_PORT, TGT_MOSI);
		} else {
			bit_clear(TGT_PORT, TGT_MOSI);
		}
		if (bit_is_set(TGT_PIN, TGT_MISO)) {
			rx |= i;
		}
		spi_clock();
		i >>= 1;
	} while (i);
	return rx;
}

static inline uint8_t spi_read(void) {
	return spi_write(0x00);
}

static uint8_t target_signature[3];
#endif
static uint8_t target_pagesize;
#define pgmmode	(target_pagesize)


static uint8_t extmode = 0;
//#define EXT_BUFSIZE	16 : move to usb910.h
#ifdef JTAG_EXTENSION
uint8_t ext_buf[JTAG_BUFSIZE+1];
#else
uint8_t ext_buf[EXT_BUFSIZE];
#endif

// tiny13   : 1e,90,07   16
//
// tiny2313 : 1e,91,0a   16
// tiny26   : 1e,91,09   16
// tiny24   : 1e,91,0b   16
// tiny25   : 1e,91,08   16
// tiny261  : 1e,91,0c   16
//
// tiny44   : 1e,92,07   32
// tiny45   : 1e,92,06   32
// tiny461  : 1e,92,08   32
//
// tiny84   : 1e,93,0c   32
// tiny85   : 1e,93,0b   32
// tiny861  : 1e,93,0d   32
// 
// mega48   : 1e,92,05   32
// mega48p  : 1e,92,0a   32
// mega48pa : 1e,92,0a   32
// mega8a   : 1e,93,07   32
// mega88   : 1e,93,0a   32
// mega88p  : 1e,93,0f   32
// mega88pa : 1e,93,0f   32
// mega8u2  : 1e,93,89   32
// 
// mega168  : 1e,94,06   64
// mega168p : 1e,94,0b   64
// mega168pa: 1e,94,0b   64
// 90usb162 : 1e,94,82   64
// mega16u2 : 1e,94,89   64
// mega16u4 : 1e,94,88   64
// 
// mega328a : 1e,95,0f   64
// mega32u2 : 1e,95,8a   64
// mega32u4 : 1e,95,87   64

#ifndef NO_AVR910
static inline void get_pagesize() {
	if (target_signature[2] <= 0x91) { // 1K or 2K
		target_pagesize = 16;
	} else if (target_signature[2] <= 0x93) { // 4K or 8K
		target_pagesize = 32;
	} else if (target_signature[2] <= 0x95) { // 16K or 32K
		target_pagesize = 64;
	} else {
		target_pagesize = 128;
	}
}

static const uint8_t twd_flash = 10;
static const uint8_t twd_eeprom = 10;
static const uint8_t twd_erase = 10;
#endif

static inline void init_spi() {
	TGT_DDR &=  ~((1<<TGT_MISO));
	TGT_PORT |=  ((1<<TGT_RESET));
	TGT_PORT &=  ~((1<<TGT_SCK) | (1<<TGT_MOSI) |(1<<TGT_MISO) );
	TGT_DDR |=    ((1<<TGT_SCK) | (1<<TGT_MOSI) |(1<<TGT_RESET) );
#ifdef USE_USI_SPI
	USICR = (1<<USIWM0); // enable USI as SPI Master
#endif
}

static inline void term_spi() {
#ifdef USE_USI_SPI
	USICR = 0; // disable USI
#endif
	TGT_DDR &=   ~((1<<TGT_SCK) | (1<<TGT_MOSI) | (1<<TGT_RESET) |(1<<TGT_MISO));
	TGT_PORT &=  ~((1<<TGT_SCK) | (1<<TGT_MOSI) | (1<<TGT_RESET) |(1<<TGT_MISO));
}

#ifndef NO_AVR910
union {
	uint16_t word;
	struct {
		uint8_t lo;
		uint8_t hi;
	};
} addr;

static uint8_t do_avrisp4(uint8_t cmd,uint8_t d1,uint8_t d2,uint8_t d3) {
	spi_write(cmd);
	spi_write(d1);
	spi_write(d2);
	return spi_write(d3);
}

static uint8_t do_avrisp2(uint8_t cmd,uint8_t data) {
	return do_avrisp4(cmd, addr.hi, addr.lo, data);
}

static void enter_programming_mode() {
	int i;
	if (pgmmode != 0) return;

#ifdef OSC_EXTENSION
	term_osc();
#endif
	init_timeout();
        init_spi();

	for (i=0; i<5; i++) {
		bit_clear(TGT_PORT, TGT_RESET);
		mdelay(5);
		bit_set(TGT_PORT, TGT_RESET);
		mdelay(5);
		bit_clear(TGT_PORT, TGT_RESET);
		mdelay(5);

		/* enable programming */
		spi_write(0xAC);
		spi_write(0x53);
		if (spi_write(0x00) != 0x53) {
			spi_write(0x00);
			continue;
		}
		spi_write(0x00);
		for (i =0 ; i < 3; i++) {
			target_signature[i] = do_avrisp4(0x30,0x00,i,0x00);
		}
		if (target_signature[0] == 0x1e) {
			get_pagesize();
			break;
		}
	}
	mdelay(5);
}

static inline void leave_programming_mode() {
        term_spi();
	pgmmode = 0;
}
#endif

#ifdef JTAG_EXTENSION
#define		par_putcx_start()
#define		par_putcx	putc
#define		par_putcx_stop()
#define		par_getcx	getc
/* JTAG COMMAND format
 *
 *  +-----------+-----+-----+-----------+-----------+    +-----------+
 *  | COMMONAD  |FLAGS|BITS |   BYTES   |   DATA 1  |... |   DATA N  |
 *  +-----------+-----+-----+-----------+-----------+    +-----------+
 *       8         5     3       8
 *
 *  N : BYTES + (BITS != 0)?1:0 
 *  DATA: MSB first bit-stream
 *  DATA length : 8 * BYTES + BITS
 *
 */
#define TGT_TDI		TGT_MOSI
#define TGT_TDO		TGT_MISO
#define TGT_TCK		TGT_SCK
#define TGT_TMS		TGT_RESET

/* SET command parameter */
#define JTAG_SET_TDI	0x80
#define JTAG_SET_TMS	0x40
#define JTAG_SET_TCK	0x20
#define JTAG_SET_DELAY	0x0f

uint8_t jtag_delay = 1;
uint8_t jtag_flags;
#define JTAG_RECIEVE	0x80
#define JTAG_TMS_HIGH	0x40
#define JTAG_TDI_HIGH	0x20
#define JTAG_USE_DELAY	0x10
#define JTAG_BITS	0x07
#define JTAG_CONST_TMS	0x01 /* interval */

#define jtag_buf ext_buf
uint8_t jtag_bytes;
uint8_t jtag_bits;
uint8_t jtag_bitoff;

static inline void jtag_set(uint8_t flags) {
	if (flags & JTAG_SET_TDI) {
#ifdef USE_USI_SPI
		bit_set(USIDR, 7);
#else
		bit_set(TGT_PORT, TGT_TDI);
#endif
	} else {
#ifdef USE_USI_SPI
		bit_clear(USIDR, 7);
#else
		bit_clear(TGT_PORT, TGT_TDI);
#endif
	}
	if (flags & JTAG_SET_TMS) {
		bit_set(TGT_PORT, TGT_TMS);
	} else {
		bit_clear(TGT_PORT, TGT_TMS);
	}
	if (flags & JTAG_SET_TCK) {
		bit_set(TGT_PORT, TGT_TCK);
	} else {
		bit_clear(TGT_PORT, TGT_TCK);
	}
	jtag_delay = (uint8_t)(F_CPU/4000000) * (flags & JTAG_SET_DELAY) + 1;
	             /* 2 @8Mhz, 3@12Mhz */
}

static inline void jtag_prepare(uint8_t flags) {
	jtag_flags = flags;
	if (flags & JTAG_CONST_TMS) {
		if (flags & JTAG_TMS_HIGH) {
			bit_set(TGT_PORT, TGT_TMS);
		} else {
			bit_clear(TGT_PORT, TGT_TMS);
		}
	}
	jtag_bytes = 0;
	jtag_bits = 0;
	jtag_bitoff = 0x80;
}

static inline void jtag_send_8(uint8_t flags, uint8_t data, uint8_t bits) {
	uint8_t rcv = 0;
#ifdef USE_USI_SPI
	uint8_t v1,v2;
#ifdef USI_TEST
	/* F_CPU/6 bitclock(max) */
	uint8_t i;
	v1 = (1<<USIWM0)|(1<<USITC);
	v2 = (1<<USIWM0)|(1<<USITC)|(1<<USICLK);
	USIDR = data;
	for (i=0; i<8; i++) {
		USICR = v1;
		nop();
		nop();
		USICR = v2;
	}
	rcv = USIDR;
#else
	/* F_CPU/2 bitclock */
	v1 = (1<<USIWM0)|(1<<USITC);
	v2 = (1<<USIWM0)|(1<<USITC)|(1<<USICLK);
	USIDR = data;
	USICR = v1;
	USICR = v2;
	USICR = v1;
	USICR = v2;
	USICR = v1;
	USICR = v2;
	USICR = v1;
	USICR = v2;
	USICR = v1;
	USICR = v2;
	USICR = v1;
	USICR = v2;
	USICR = v1;
	USICR = v2;
	USICR = v1;
	USICR = v2;
	rcv = USIDR;
#endif
#else
	uint8_t i;
	for (i=0; i<8; i++) {
		rcv <<= 1;
		if (data & 0x80) {
			bit_set(TGT_PORT, TGT_TDI);
		} else {
			bit_clear(TGT_PORT, TGT_TDI);
		}
		data <<= 1;
		bit_set(TGT_PORT, TGT_TCK);
		if (bit_is_set(TGT_PIN, TGT_TDO)) {
			rcv |= 1;
		}
		nop();
		bit_clear(TGT_PORT, TGT_TCK);
	}
#endif
	if (flags & JTAG_RECIEVE) {
		jtag_buf[jtag_bytes++] = rcv;
	}
}

static void jtag_send_d(uint8_t flags, uint8_t data, uint8_t bits) {
	uint8_t i,delay;

	delay = jtag_delay;
	for (i=0; i<bits; i++) {
		if (data & 0x80) {
#ifdef USE_USI_SPI
			bit_set(USIDR, 7);
#else
			bit_set(TGT_PORT, TGT_TDI);
#endif
		} else {
#ifdef USE_USI_SPI
			bit_clear(USIDR, 7);
#else
			bit_clear(TGT_PORT, TGT_TDI);
#endif
		}
		data <<= 1;
		if (!(flags & JTAG_CONST_TMS)) {
			if (data & 0x80) {
				bit_set(TGT_PORT, TGT_TMS);
			} else {
				bit_clear(TGT_PORT, TGT_TMS);
			}
			data <<= 1;
		}
		_delay_loop_1(delay);
		bit_set(TGT_PORT, TGT_TCK);
		if (flags & JTAG_RECIEVE) {
			if (bit_is_set(TGT_PIN, TGT_TDO)) {
				jtag_buf[jtag_bytes] |= jtag_bitoff;
			} else {
				jtag_buf[jtag_bytes] &= ~jtag_bitoff;
			}
			jtag_bits++;
			jtag_bitoff >>= 1;
			if (jtag_bitoff == 0) {
				jtag_bitoff = 0x80;
				jtag_bytes++;
			}
		}
		_delay_loop_1(delay);
		bit_clear(TGT_PORT, TGT_TCK);
	}
}
#endif

int main(void)
{
    uint8_t ch;
#ifndef NO_AVR910
    uint8_t d1,d2,d3;
#endif
#if defined(AVR109_EXTENSION)
    uint16_t size;
#endif


#ifdef F_SCALER
    /* set system_clock pre scaler 1/8 => 1/1 */
    CLKPR = (1<<CLKPCE);
    CLKPR = F_SCALER;
#endif

#if defined(__CDC162__) 
    // bit_set(REGCR, REGDIS); // disable internal regulator 

    ch = 20;
    while (--ch) {
        _delay_ms(1);
    }
#endif
    //wdt_enable(WDTO_1S);
    usbcdc_init();

    sei();

    for(;;) {
	wdt_reset();
	poll();

	ch = getc();

	if (ch == 'S') { // get Software
#if defined(NO_AVR910) && defined(JTAG_EXTENSION)
		putc('S');
		putc('E');
		putc('R');
		putc('J');
		putc('T');
		putc('A');
		putc('G');
#else
		putc('U');
		putc('S');
		putc('B');
		putc('9');
		putc('1');
		putc('0');
  #if defined(__AVR_ATtiny861__)
		putc('A');
  #else
		putc('B');
  #endif
#endif
#ifndef NO_AVR910
	} else if (ch == 'p') { // get Programmer Type
		putc('S'); // Serial
#endif
	} else if (ch == 'V' || ch == 'v') { // get Version 'V':sw 'v':hw
		extmode = 0;
		pgmmode = 0;
#if defined(NO_AVR910) && defined(JTAG_EXTENSION)
		putc('1'); // Revision
 #if defined(ADC_EXTENSION) && defined(FCNT_EXTENSION) && defined(OSC_EXTENSION)
		putc('3');
 #elif defined(ADC_EXTENSION) && defined(FCNT_EXTENSION)
		putc('2');
 #elif defined(ADC_EXTENSION)
		putc('1');
 #else
		putc('0');
 #endif
#else
		putc('2');
		putc('4');
#endif
#ifndef NO_AVR910
	} else if (ch == 'a') { // Report autoincrement address
		extmode = 0;
		pgmmode = 0;
		putc('Y'); // Yes!
	} else if (ch == 'A') { // Set Write Address
		addr.hi=getc();
		addr.lo=getc();
		putc('\r'); // ACK
	} else if (ch == 't') { // get supported Device Types
		putc(0x20); // dummy 
		putc(0);
	} else if (ch == 'T') { // set Device Types: ignore but ACK
		getc();
		putc('\r');
#if 0
	} else if (ch == 'x' || ch == 'y') { // Set/Clear LED
		putc('\r'); // ACK
#endif
	} else if (ch == 'P') { // Enter Programming mode
		enter_programming_mode();
		putc(pgmmode?'\r':'?');
	} else if (ch == 'L') { // Leave Programming mode, NOP but ACK
		leave_programming_mode();
		putc('\r');
#endif
#ifdef AVR109_EXTENSION
	} else if (ch == 'b') {// Return BufferSize
		enter_programming_mode(); // need target_pagesize
		if (pgmmode) {
			putc('Y');
			putc((target_pagesize & 0x80)?1:0);
			putc(target_pagesize<<1);
		} else {
			putc('?');
		}
#endif
	} else if ((ch == 0x1b) || (ch == '\n')) { // ESC,\n: nop
#ifndef NO_AVR910
	} else if (pgmmode) {
	    if (ch == 's') { // Return Signature Byte
		putc(target_signature[2]); // part_number
		putc(target_signature[1]); // part_family
		putc(target_signature[0]); // vendor code
	    } else if (ch == 'e') { // Chip Erase
		do_avrisp4(0xAC,0x80,0x00,0x00);
		mdelay(twd_erase);
		putc('\r'); // ACK
#ifndef OFF_LEGACY
	    } else if (ch == 'c') { // Write program memory, low byte
		do_avrisp2(0x40,getc()); 
		putc('\r'); // ACK
	    } else if (ch == 'C') { // Write program memory, high byte
		do_avrisp2(0x48,getc()); 
		addr.word++;
		putc('\r'); // ACK
	    } else if (ch == 'm') { // Issue Page Write
		do_avrisp2(0x4C,0x00); 
		mdelay(twd_flash);
		putc('\r'); // ACK
	    } else if (ch == 'R') { // Read program memory
		putc(do_avrisp2(0x28,0x00)); // Read high byte
		putc(do_avrisp2(0x20,0x00)); // Read low byte
		addr.word++;
	    } else if (ch == 'D') { // Write data memory
		do_avrisp2(0xC0,getc());
		mdelay(twd_eeprom);
		addr.word++;
		putc('\r'); // ACK
	    } else if (ch == 'd') { // Read data memory
		putc(do_avrisp2(0xA0,0x00));
		addr.word++;
#endif
	    } else if (ch == 'l') { // Write lock bits
		do_avrisp4(0xAC,0xE0,0x00,getc());
		mdelay(twd_flash);
		putc('\r'); // ACK
	    } else if (ch == '.') { // New universal command
		d1 = getc();
		d2 = getc();
		d3 = getc();
		putc(do_avrisp4(d1,d2,d3,getc()));
		putc('\r'); // ACK
#ifdef AVR109_EXTENSION
	    } else if (ch == 'r') { // Read Lock Bits
		putc(do_avrisp4(0x58,0x00,0x00,0x00));
	    } else if (ch == 'F') { // Read L-Fuses
		putc(do_avrisp4(0x50,0x00,0x00,0x00));
	    } else if (ch == 'N') { // Read H-Fuses
		putc(do_avrisp4(0x58,0x08,0x00,0x00));
	    } else if (ch == 'Q') { // Read EXT-Fuses
		putc(do_avrisp4(0x58,0x08,0x00,0x00));
	    } else if (ch == 'g') { // Read memory block
		size = getc() << 8;	// high byte of buffersize
		size |= getc();	// low byte
		ch = getc();   // memory type

		if (ch == 'E') { // EEPROM

		    while (size--) {
			putc(do_avrisp2(0xA0,0x00));
			addr.word++;
		    }

	        } else if (!(addr.lo & 1) && !(size & 1)) { // FLASH
		    addr.word >>= 1;
		    size >>= 1;
		    while (size--) {
			putc(do_avrisp2(0x20,0x00)); // Read low byte
			putc(do_avrisp2(0x28,0x00)); // Read high byte
			addr.word++;
		    }
		    addr.word <<= 1;
		} else {
		    while (size--) {
		   	 putc(0);
		    }
		}

	    } else if (ch == 'B') { // Programm with buffer

		size = getc() << 8;	// high byte of buffersize
		size |= getc();	// low byte
		ch = getc();   // memory type

		if (ch == 'E')	{ // EEPROM
		    while (size--) {
			do_avrisp2(0xC1,getc());
			if ((addr.lo & 0x3) == 0x3) {
			    do_avrisp2(0xC2,getc());
			    mdelay(twd_eeprom);
		 	}
			addr.word++;
		    }
		    putc('\r'); // ACK

		} else if (ch == 'F' && !(addr.lo & ((target_pagesize<<1) -1))
				&& !(size & 1) ) { // FLASH 
		    addr.word >>= 1;
		    size >>= 1;
		    while (size--) {
			do_avrisp2(0x40,getc()); // Write low byte
			do_avrisp2(0x48,getc()); // Write high byte
			if (size == 0) {
			    do_avrisp2(0x4C,0x00); // Issue Page Write
			    mdelay(twd_flash);
			}
		        addr.word++;
		    }
		    addr.word <<= 1;
		    putc('\r'); // ACK

		} else {
		    putc(0); // Unsupported memory type or wrong params
		}
#endif
	    } else {
		putc(0); // Unsupported memory type or wrong params
	    }
#endif
#ifdef JTAG_EXTENSION
	} else if (extmode == 2) {
	    uint8_t bytes,bits,flags;
	    uint8_t i,r;
	   if (ch == 'r') { // Request Recieved Data
		getc(); getc(); // skip header
		flags = jtag_bits & JTAG_BITS;
		bytes = jtag_bytes;
		if (flags) bytes++;
		putc('R');
		putc(flags);
		putc(jtag_bytes);
		par_putcx_start();
		for (i=0 ; i< bytes; i++) {
			par_putcx(jtag_buf[i]);
		}
		par_putcx_stop();
		jtag_bits = 0;
		jtag_bytes = 0;
	   } else if (ch == 's') { //  Set Port
		getc(); getc(); // skip header
		jtag_set(getc());
	   } else if (ch == 'd') { // Put TDI Stream
		bits = getc();
		bytes = getc();
		flags = (bits & ~JTAG_BITS) | JTAG_CONST_TMS;
		bits &= JTAG_BITS;
		jtag_prepare(flags);
		if (flags & JTAG_USE_DELAY) {
			for (i=0; i<bytes; i++) {
				jtag_send_d(flags, getc(), 8);
			}
		} else {
			for (i=0; i<bytes; i++) {
				jtag_send_8(flags, par_getcx(), 8);
			}
		}
		if (bits) {
			jtag_send_d(flags, getc(), bits);
		}
	   } else if (ch == 'D') { // Put TDI+TMS Stream
		bits = getc();
		bytes = getc();
		flags = (bits & ~JTAG_BITS);
		bits &= JTAG_BITS;
		for (i=0; i<bytes; i++) {
			jtag_send_d(flags, getc(), 4);
		}
		bits >>= 1;
		if (bits) {
			jtag_send_d(flags, getc(), bits);
		}
	   } else if (ch == 'c') { // Get TDO Stream
		bits = getc();
		bytes = getc();
		flags = (bits & ~JTAG_BITS) | JTAG_RECIEVE;
		bits &= JTAG_BITS;
		jtag_prepare(flags);
		r = (flags & JTAG_TDI_HIGH)? 0xff: 0;
		if (flags & JTAG_USE_DELAY) {
			for (i=0; i<bytes; i++) {
				jtag_send_d(flags, r, 8);
			}
		} else {
			for (i=0; i<bytes; i++) {
				jtag_send_8(flags, r, 8);
			}
		}
		if (bits) {
			jtag_send_d(flags, r, bits);
		}
	   }
#endif
	} else if (extmode == 1) {
	    uint8_t bytes,i;
	    uint8_t dev;
	    int8_t r;
	    r = -1; // device is not existed
	    dev = getc();
	    bytes = getc();
	    if (ch == 's' ) { // Send
		for (i=0; i<bytes && i< EXT_BUFSIZE; i++) {
		    ext_buf[i] = getc();
		}
		for (; i<bytes; i++) {
			getc();
		}
		bytes = i;
		if (dev == DEV_SYS) {
#ifdef ADC_EXTENSION
		} else if (dev == DEV_ADC) {
		    r = adc_send(bytes);
#endif
#ifdef OSC_EXTENSION
		} else if (dev == DEV_OSC) {
		    r = osc_send(bytes);
#endif
#ifdef FCNT_EXTENSION
		} else if (dev == DEV_FCNT) {
		    r = fcnt_send(bytes);
#endif
#ifndef NO_DEV_ECHO
	        } else if (dev == DEV_ECHO) {
		    r = bytes;
#endif
#ifdef I2C_EXTENSION
		} else if (DEV_I2C <= dev && dev <= 0x7f) {
		    r = i2c_send(dev, bytes);
#endif
		}
		putc(r);
	    } else if (ch == 'r') { // Recv
		 if (bytes > EXT_BUFSIZE)
			 bytes = EXT_BUFSIZE;
		if (dev == DEV_SYS) {
#ifdef ADC_EXTENSION
	        } else if (dev == DEV_ADC) {
		    r = adc_recv(bytes);
#endif
#ifdef OSC_EXTENSION
	        } else if (dev == DEV_OSC) {
		    r = osc_recv(bytes);
#endif
#ifdef FCNT_EXTENSION
	        } else if (dev == DEV_FCNT) {
		    r = fcnt_recv(bytes);
#endif
#ifndef NO_DEV_ECHO
	        } else if (dev == DEV_ECHO) {
		    r = bytes;
#endif
#ifdef I2C_EXTENSION
		} else if (DEV_I2C <= dev && dev <= 0x7f) {
		    r = i2c_recv(dev, bytes);
#endif
		}
		putc(r);
		for (i=0; i<r; i++) {
		    putc(ext_buf[i]);
	        }
	   } else {
		putc(r);
	   }
	} else if (ch == 'e') { // Enter Extent mode 
		extmode = 1;
#ifdef I2C_EXTENSION
		init_i2c();
#endif
#ifdef ADC_EXTENSION
		init_adc();
#endif
#ifdef OSC_EXTENSION
		init_osc();
#endif

		putc('Y');
#ifdef JTAG_EXTENSION
	} else if (ch == 'j') { // Enter Extent mode 2 (JTAG)
		extmode = 2;
		jtag_bytes = 0;
		jtag_bits = 0;
		init_spi();
		putc('Y');
#endif
	} else {
		putc('?');
	}
    }
    return 0;
}

#if defined(__CDC162__) 
void callback_set_baud(long baud) {
	//Baud = baud;
}
#endif

