/*
 * isp.c - part of USBasp
 *
 * Autor..........: Thomas Fischl <tfischl@gmx.de>
 * Description....: Provides functions for communication/programming
 *                  over ISP interface
 * Licence........: GNU GPL v2 (see Readme.txt)
 * Creation Date..: 2005-02-23
 * Last change....: 2009-02-28
 */

#include <avr/io.h>
#include <avr/wdt.h>
#include <isp.h>
#include <clock.h>
#include <aspconfig.h>

#define spiHWdisable() SPCR = 0

static uchar sck_sw_delay;
static uchar (*ispTransmit)(uchar);

static void ispDelay() {

	uint8_t starttime = TIMERVALUE;
	while ((uint8_t) (TIMERVALUE - starttime) < sck_sw_delay) {
		wdt_reset();
	}
}

#ifdef ENABLE_ISP1
static uchar sck_spcr;
static uchar sck_spsr;
/* pointer to sw or hw transmit function */

/* read an write a byte from isp using software (slow) */
static uchar ispTransmit_sw(uchar send_byte);

/* read an write a byte from isp using hardware (fast) */
static uchar ispTransmit_hw(uchar send_byte);

static void spiHWenable() {
	SPCR = sck_spcr;
	SPSR = sck_spsr;
}

static void ispSetSCKOption(uchar option) {

	if (option == USBASP_ISP_SCK_AUTO)
		option = USBASP_ISP_SCK_250;

	if (option >= USBASP_ISP_SCK_62_5) {
		ispTransmit = ispTransmit_hw;
		sck_spsr = 0;
		sck_sw_delay = 3;

		switch (option) {
#if F_CPU == 8000000UL
		case USBASP_ISP_SCK_4000:
			/* enable SPI, master, 4.0MHz, XTAL/2 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (0 << SPR0);
			sck_spsr = (1 << SPI2X);
			break;
		case USBASP_ISP_SCK_2000:
			/* enable SPI, master, 2.0MHz, XTAL/4 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (0 << SPR0);
			break;
		case USBASP_ISP_SCK_1000:
			/* enable SPI, master, 1.0MHz, XTAL/8 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (1 << SPR0);
			sck_spsr = (1 << SPI2X);
			break;
		case USBASP_ISP_SCK_500:
			/* enable SPI, master, 500kHz, XTAL/16 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (1 << SPR0);
			break;
		case USBASP_ISP_SCK_250:
		default:
			/* enable SPI, master, 250kHz, XTAL/32 (default) */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (2 << SPR0);
			sck_spsr = (1 << SPI2X);
			break;
		case USBASP_ISP_SCK_125:
			/* enable SPI, master, 125kHz XTAL/64 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (2 << SPR0);
			break;
		case USBASP_ISP_SCK_62_5:
			/* enable SPI, master, 62.5 kHz XTAL/128 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (3 << SPR0);
			sck_spsr = (1 << SPI2X);
			break;
		}
#elif F_CPU == 16000000UL
		case USBASP_ISP_SCK_4000:
			/* enable SPI, master, 4.0MHz, XTAL/4 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (0 << SPR0);
			break;
		case USBASP_ISP_SCK_2000:
			/* enable SPI, master, 2.0MHz, XTAL/8 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (1 << SPR0);
			sck_spsr = (1 << SPI2X);
			break;
		case USBASP_ISP_SCK_1000:
			/* enable SPI, master, 1.0MHz, XTAL/16 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (1 << SPR0);
			break;
		case USBASP_ISP_SCK_500:
			/* enable SPI, master, 500kHz, XTAL/32 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (2 << SPR0);
			sck_spsr = (1 << SPI2X);
			break;
		case USBASP_ISP_SCK_250:
		default:
			/* enable SPI, master, 250kHz, XTAL/32 (default) */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (2 << SPR0);
			break;
		case USBASP_ISP_SCK_125:
			/* enable SPI, master, 125kHz XTAL/64 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (3 << SPR0);
			sck_spsr = (1 << SPI2X);
			break;
		case USBASP_ISP_SCK_62_5:
			/* enable SPI, master, 62.5 kHz XTAL/128 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (3 << SPR0);
			break;
		}
#else
		case USBASP_ISP_SCK_6000:
			/* enable SPI, master, 6MHz, XTAL/2 */
			sck_spcr = (1 << SPE) | (1 << MSTR);
			sck_spsr = (1 << SPI2X);
		case USBASP_ISP_SCK_3000:
			/* enable SPI, master, 3MHz, XTAL/4 */
			sck_spcr = (1 << SPE) | (1 << MSTR);
			break;
		case USBASP_ISP_SCK_1500:
			/* enable SPI, master, 1.5MHz, XTAL/8 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (1 << SPR0);
			sck_spsr = (1 << SPI2X);
		case USBASP_ISP_SCK_750:
			/* enable SPI, master, 750kHz, XTAL/16 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (1 << SPR0);
			break;
		case USBASP_ISP_SCK_375:
		default:
			/* enable SPI, master, 375kHz, XTAL/32 (default) */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (2 << SPR0);
			sck_spsr = (1 << SPI2X);
			break;
		case USBASP_ISP_SCK_187_5:
			/* enable SPI, master, 187.5kHz XTAL/64 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (2 << SPR0);
			break;
		case USBASP_ISP_SCK_93_75:
			/* enable SPI, master, 93.75kHz XTAL/128 */
			sck_spcr = (1 << SPE) | (1 << MSTR) | (3 << SPR0);
			break;
		}
#endif

	} else {
		ispTransmit = ispTransmit_sw;
		switch (option) {

		case USBASP_ISP_SCK_32:
			sck_sw_delay = 3;

			break;
		case USBASP_ISP_SCK_16:
			sck_sw_delay = 6;

			break;
		case USBASP_ISP_SCK_8:
			sck_sw_delay = 12;

			break;
		case USBASP_ISP_SCK_4:
			sck_sw_delay = 24;

			break;
		case USBASP_ISP_SCK_2:
			sck_sw_delay = 48;

			break;
		case USBASP_ISP_SCK_1:
			sck_sw_delay = 96;

			break;
		case USBASP_ISP_SCK_0_5:
			sck_sw_delay = 192;

			break;
		}
	}
}

static void ispDisconnect() {

	/* set all ISP pins inputs */
	ISP_DDR &= ~((1 << ISP_RST) | (1 << ISP_SCK) | (1 << ISP_MOSI));
	/* switch pullups off */
	ISP_OUT &= ~((1 << ISP_RST) | (1 << ISP_SCK) | (1 << ISP_MOSI));

	/* disable hardware SPI */
	spiHWdisable();
}

static uchar ispTransmit_sw(uchar send_byte) {

	uchar rec_byte = 0;
	uchar i;
	for (i = 0; i < 8; i++) {

		/* set MSB to MOSI-pin */
		if ((send_byte & 0x80) != 0) {
			ISP_OUT |= (1 << ISP_MOSI); /* MOSI high */
		} else {
			ISP_OUT &= ~(1 << ISP_MOSI); /* MOSI low */
		}
		/* shift to next bit */
		send_byte = send_byte << 1;

		/* receive data */
		rec_byte = rec_byte << 1;
		if ((ISP_IN & (1 << ISP_MISO)) != 0) {
			rec_byte++;
		}

		/* pulse SCK */
		ISP_OUT |= (1 << ISP_SCK); /* SCK high */
		ispDelay();
		ISP_OUT &= ~(1 << ISP_SCK); /* SCK low */
		ispDelay();
	}

	return rec_byte;
}

static uchar ispTransmit_hw(uchar send_byte) {
	SPDR = send_byte;

	while (!(SPSR & (1 << SPIF)))
		;
	return SPDR;
}

static uchar ispEnterProgrammingMode() {
	uchar check;
	uchar count = 32;

	while (count--) {
		ispTransmit(0xAC);
		ispTransmit(0x53);
		check = ispTransmit(0);
		ispTransmit(0);

		if (check == 0x53) {
			return 0;
		}

		spiHWdisable();

		/* pulse SCK */
		ISP_OUT |= (1 << ISP_SCK); /* SCK high */
		ispDelay();
		ISP_OUT &= ~(1 << ISP_SCK); /* SCK low */
		ispDelay();

		if (ispTransmit == ispTransmit_hw) {
			spiHWenable();
		}

	}

	return 1; /* error: device dosn't answer */
}
#endif

#ifdef ENABLE_ISP2
static uchar isp2_sck;
 
static void isp2SetSCKOption(uchar option) {
	if ((option == USBASP_ISP_SCK_AUTO) || ( option > USBASP_ISP_SCK_8000)) 
		option = USBASP_ISP_SCK_125;
#if F_CPU == 8000000UL
	if ( option > USBASP_ISP_SCK_4000) 
		option = USBASP_ISP_SCK_4000;
	UBRR1 = (1 << (14 - option)) - 1;
#elif F_CPU == 16000000UL
	UBRR1 = (1 << (15 - option)) - 1;
#elif F_CPU == 12000000UL
	if ( option > USBASP_ISP_SCK_2000) 
		option = USBASP_ISP_SCK_2000;
	UBRR1 = (3 << (13 - option)) - 1;
#endif
	sck_sw_delay = 3;
	switch (option) {
	case USBASP_ISP_SCK_32:
		sck_sw_delay = 3;
		break;
	case USBASP_ISP_SCK_16:
		sck_sw_delay = 6;
		break;
	case USBASP_ISP_SCK_8:
		sck_sw_delay = 12;
		break;
	case USBASP_ISP_SCK_4:
		sck_sw_delay = 24;
		break;
	case USBASP_ISP_SCK_2:
		sck_sw_delay = 48;
		break;
	case USBASP_ISP_SCK_1:
		sck_sw_delay = 96;
		break;
	case USBASP_ISP_SCK_0_5:
		sck_sw_delay = 192;
		break;
	}
	isp2_sck = option;
}

static void isp2HWenable() {
	UCSR1C = (3<<UMSEL10) | (0<<UCPOL1);
	UCSR1B = (1<<RXEN1) | (1<<TXEN1);
}

static void isp2HWdisable() {
	UCSR1B = 0;
	UCSR1C = 0;
}

static void isp2Disconnect() {

	/* set all ISP pins inputs */
	ISP2_DDR &= ~((1 << ISP2_RST) | (1 << ISP2_SCK) | (1 << ISP2_MOSI));
	/* switch pullups off */
	ISP2_OUT &= ~((1 << ISP2_RST) | (1 << ISP2_SCK) | (1 << ISP2_MOSI));

	/* disable hardware SPI */
	isp2HWdisable();
}

static uchar isp2Transmit(uchar send_byte) {
	UDR1 = send_byte;

	while (!(UCSR1A & (1 << RXC1)))
		;
	return UDR1;
}

static uchar isp2EnterProgrammingMode() {
	uchar check;
	uchar count = 32;

	while (count--) {
		isp2Transmit(0xAC);
		isp2Transmit(0x53);
		check = isp2Transmit(0);
		isp2Transmit(0);

		if (check == 0x53) {
			return 0;
		}

		isp2HWdisable();

		/* pulse SCK */
		ISP2_OUT |= (1 << ISP2_SCK); /* SCK high */
		ispDelay();
		ISP2_OUT &= ~(1 << ISP2_SCK); /* SCK low */
		ispDelay();

		isp2HWenable();
	}

	return 1; /* error: device dosn't answer */
}
#endif

/* read and write 4 bytes from isp */
static void ispTransmit4(uchar *out, uchar *in) {
	out[0] = ispTransmit(in[0]);
	out[1] = ispTransmit(in[1]);
	out[2] = ispTransmit(in[2]);
	out[3] = ispTransmit(in[3]);
}
static uchar ispReadFlash(uint32_t address) {
	ispTransmit(0x20 | ((address & 1) << 3));
	ispTransmit(address >> 9);
	ispTransmit(address >> 1);
	return ispTransmit(0);
}

static uchar ispWriteFlash(uint32_t address, uchar data, uchar pollmode) {

	/* 0xFF is value after chip erase, so skip programming
	 if (data == 0xFF) {
	 return 0;
	 }
	 */

	ispTransmit(0x40 | ((address & 1) << 3));
	ispTransmit(address >> 9);
	ispTransmit(address >> 1);
	ispTransmit(data);

	if (pollmode == 0)
		return 0;

	if (data == 0x7F) {
		clockWait(15); /* wait 4,8 ms */
		return 0;
	} else {

		/* polling flash */
		uchar retries = 30;
		uint8_t starttime = TIMERVALUE;
		while (retries != 0) {
			if (ispReadFlash(address) != 0x7F) {
				return 0;
			};

			if ((uint8_t) (TIMERVALUE - starttime) > CLOCK_T_320us) {
				starttime = TIMERVALUE;
				retries--;
			}

		}
		return 1; /* error */
	}

}

static uchar ispFlushPage(unsigned long address, uchar pollvalue) {
	ispTransmit(0x4C);
	ispTransmit(address >> 9);
	ispTransmit(address >> 1);
	ispTransmit(0);

	if (pollvalue == 0xFF) {
		clockWait(15);
		return 0;
	} else {

		/* polling flash */
		uchar retries = 30;
		uint8_t starttime = TIMERVALUE;

		while (retries != 0) {
			if (ispReadFlash(address) != 0xFF) {
				return 0;
			};

			if ((uint8_t) (TIMERVALUE - starttime) > CLOCK_T_320us) {
				starttime = TIMERVALUE;
				retries--;
			}

		}

		return 1; /* error */
	}

}

static uchar ispReadEEPROM(uint16_t address) {
	ispTransmit(0xA0);
	ispTransmit(address >> 8);
	ispTransmit(address);
	return ispTransmit(0);
}

static uchar ispWriteEEPROM(uint16_t address, uchar data) {

	ispTransmit(0xC0);
	ispTransmit(address >> 8);
	ispTransmit(address);
	ispTransmit(data);

	clockWait(30); // wait 9,6 ms

	return 0;
}

void ispConnect(struct pgm_ops *ops, uchar sck_opt, uchar chan) {
#if defined(ENABLE_ISP1) && !defined(ENABLE_ISP2)
	chan = 0;
#elif !defined(ENABLE_ISP1) && defined(ENABLE_ISP2)
	chan = 1;
#endif
	if (chan == 1) {
#ifdef ENABLE_ISP2
		ispTransmit = isp2Transmit;
		isp2HWdisable();
		isp2SetSCKOption(sck_opt);

		/* all ISP pins are inputs before */
		/* now set output pins */
		ISP2_DDR |= (1 << ISP2_RST) | (1 << ISP2_SCK) | (1 << ISP2_MOSI);

		/* reset device */
		ISP2_OUT &= ~(1 << ISP2_RST); /* RST low */
		ISP2_OUT &= ~(1 << ISP2_SCK); /* SCK low */

		/* positive reset pulse > 2 SCK (target) */
		ispDelay();
		ISP2_OUT |= (1 << ISP2_RST); /* RST high */
		ispDelay();
		ISP2_OUT &= ~(1 << ISP2_RST); /* RST low */

		isp2HWenable();
		ops->pgm_disconnect   = isp2Disconnect;
		ops->pgm_enable       = isp2EnterProgrammingMode;
#else
		return;
#endif
	} else {
#ifdef ENABLE_ISP1
		spiHWdisable();
		ispSetSCKOption(sck_opt);

		/* all ISP pins are inputs before */
		/* now set output pins */
		ISP_DDR |= (1 << ISP_RST) | (1 << ISP_SCK) | (1 << ISP_MOSI);

		/* reset device */
		ISP_OUT &= ~(1 << ISP_RST); /* RST low */
		ISP_OUT &= ~(1 << ISP_SCK); /* SCK low */

		/* positive reset pulse > 2 SCK (target) */
		ispDelay();
		ISP_OUT |= (1 << ISP_RST); /* RST high */
		ispDelay();
		ISP_OUT &= ~(1 << ISP_RST); /* RST low */

		if (ispTransmit == ispTransmit_hw) {
			spiHWenable();
		}
		ops->pgm_disconnect   = ispDisconnect;
		ops->pgm_enable       = ispEnterProgrammingMode;
#else
		return;
#endif
	}
	ops->pgm_transmit     = ispTransmit4;
	ops->pgm_read_space0  = ispReadEEPROM;
	ops->pgm_write_space0 = ispWriteEEPROM;
	ops->pgm_read_space1  = ispReadFlash;
	ops->pgm_write_space1 = ispWriteFlash;
	ops->pgm_close_page   = ispFlushPage;
};
